from typing import Any, Dict

# This is a placeholder for JSONType until it gets officially support.
# 5/2020 -- waiting on mypy support for recursion.
# https://github.com/python/typing/issues/182
JSONType = Dict[str, Any]
